# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Alexander Shorin
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

__version_info__ = (1, 0, 0)
__version__ = '.'.join(map(str, __version_info__))
